#include "esd.h"

#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <string.h>
#include <fcntl.h>

#define BUFFERSIZE (ESD_BUF_SIZE)

int finished=0;
int signaled=1;

int freq=ESD_DEFAULT_RATE;

char *esd_host = NULL;

void sighandle_zombies(int i) {
  waitpid(-1,NULL,WNOHANG);
  finished++;   
}

void MP3Init(void) {
  signal(SIGCHLD,sighandle_zombies);  
}  

void MP3Host(char *s)
{
  esd_host = malloc(strlen(s)+2);
  strcpy(esd_host, s);
}

void MP3Frequency(int f)
{
  freq=f;
}

int MP3Finished(void) {
  if (finished) {
    finished--;
    return 1;
  } else return 0;      
}

void WaitOnChild(void) {
  while (!signaled) sleep(1); // waiting for signal
}

int esd_get_stream_id(int esd,char *name) {
  int q;
  esd_info_t *allinfo;
  esd_player_info_t *plinfo;
  int retries=10;
  
  allinfo=NULL;
  do {
    if (allinfo!=NULL) esd_free_all_info(allinfo);
    allinfo = esd_get_all_info(esd);
    plinfo = allinfo->player_list;

    retries--;
    if (!retries) break;
  } while (plinfo==NULL);    
  while (plinfo!=NULL) {
    if (strcmp(plinfo->name, name)==0) {
       q= plinfo->source_id;
       esd_free_all_info(allinfo);
       return q;
    }
    plinfo = plinfo->next;
  }
  
  esd_free_all_info(allinfo);
  return -1;
}

char MP3IsPlaying(int id) {
  int sid,esd;
  char s[100];
  
  WaitOnChild();
  sprintf(s,"MP3[%d]",id);
  esd = esd_open_sound(esd_host);
  sid = esd_get_stream_id(esd,s);      
  esd_close(esd);
  return (sid!=-1);    
}

int MP3VolumeNow(int id, int l, int r) {
  int sid,esd;
  char s[100];
  
  sprintf(s,"MP3[%d]",id);
  esd = esd_open_sound(esd_host);
  sid = esd_get_stream_id(esd,s);
  esd_set_stream_pan(esd,sid,l,r);    
  esd_close(esd);
  return 0;    
}

int MP3Volume(int id, int l, int r) {
  WaitOnChild();
  MP3VolumeNow(id, l, r);
}  

void sighandle(int i) {
  signaled=1;
}  

int MP3Play(char *filename, int l, int r, int predelay) {
  int pid,sock,i,ppid;
  fd_set fd;
  struct timeval timeout;
  int pip[2],devnull;  
  char s[100];
  char buf[BUFFERSIZE];
  int length;
  
  signaled=0;
  signal(SIGTERM,sighandle);  
  ppid=getpid();
  if((pid=fork())==0) {
    sprintf(s,"MP3[%d]",getpid());
    sock = esd_play_stream_fallback
          ( ESD_BITS16 | ESD_STEREO | ESD_STREAM | ESD_PLAY, 
                         freq, esd_host, s );
    if (pipe(pip)) {
      fprintf(stderr,"Can't create pipe! Errno=%d\n",errno);
      close(sock);
      return 1;
    }  
    if ((pid=fork())==0) {
      signal(SIGTERM,SIG_DFL);
      dup2(pip[1],fileno(stdout));
      devnull=open("/dev/null",0);
      dup2(devnull,fileno(stderr));
      execl("/usr/bin/mpg123","/usr/bin/mpg123","-s", filename, NULL);
    } else {
      signal(SIGCHLD,SIG_DFL);
      kill(ppid,SIGTERM); // resume parent
      MP3VolumeNow(getpid(),l,r);      

      bzero(buf,BUFFERSIZE); 
// Inicializacia      
      for (i=0;i<predelay;i++)
        write( sock, buf, BUFFERSIZE );


      while (1) {        
          FD_ZERO(&fd);
          timeout.tv_sec=1;
          timeout.tv_usec=0;
          FD_SET(pip[0],&fd);
          select(pip[0]+1,&fd,NULL,NULL,&timeout);
          
          if (signaled) {
            kill(pid,SIGTERM);
            MP3VolumeNow(getpid(),0,0);
	    break;
          }          
          
          if (waitpid(-1,NULL,WNOHANG)) break;
          
          if (FD_ISSET(pip[0],&fd)) {
            length = read(pip[0], buf, BUFFERSIZE);
            write( sock, buf, length );
          }                
      }
      close( sock );
      close( pip[0] );
      execl("/bin/true","/bin/true",NULL);
    }                                     
     
  } else {
//  WaitOnChild();
    return pid;
  }    
  return 0;
}

/*
int my_system (const char *command) {
   int pid, status;

   if (command == 0)
       return 1;
   pid = fork();
   if (pid == -1)
       return -1;
   if (pid == 0) {
       char *argv[4];
       argv[0] = "sh";
       argv[1] = "-c";
       argv[2] = command;
       argv[3] = 0;
       execve("/bin/sh", argv, 0);
       exit(127);
   }
   do {
       if (waitpid(pid, &status, 0) == -1) {
                   if (errno != EINTR)
	               return -1;
       } else
           return status;
   } while(1);
}
*/

int MP3PlayLoop(char *filename, int l, int r, int predelay) {
  int pid,sock,i,ppid;
  fd_set fd;
  struct timeval timeout;
  int pip[2],devnull;  
  char s[100];
  char buf[BUFFERSIZE];
  int length;
  char cmd[128];
  
  snprintf(cmd,sizeof(cmd),"while [ 1 ]; do /usr/bin/mpg123 -s %s; done",filename);
  
  signaled=0;
  signal(SIGTERM,sighandle);  
  ppid=getpid();
  if((pid=fork())==0) {
    sprintf(s,"MP3[%d]",getpid());
    sock = esd_play_stream_fallback
          ( ESD_BITS16 | ESD_STEREO | ESD_STREAM | ESD_PLAY, 
                         freq, esd_host, s );
    if (pipe(pip)) {
      fprintf(stderr,"Can't create pipe! Errno=%d\n",errno);
      close(sock);
      return 1;
    }  
    if ((pid=fork())==0) {
      signal(SIGTERM,SIG_DFL);
      dup2(pip[1],fileno(stdout));
      devnull=open("/dev/null",0);
      dup2(devnull,fileno(stderr));
      system(cmd);
//      execl("/usr/bin/mpg123","/usr/bin/mpg123","-s", filename, NULL);
    } else {
      signal(SIGCHLD,SIG_DFL);
      kill(ppid,SIGTERM); // resume parent
      MP3VolumeNow(getpid(),l,r);      

      bzero(buf,BUFFERSIZE); 
// Inicializacia      
      for (i=0;i<predelay;i++)
        write( sock, buf, BUFFERSIZE );


      while (1) {        
          FD_ZERO(&fd);
          timeout.tv_sec=1;
          timeout.tv_usec=0;
          FD_SET(pip[0],&fd);
          select(pip[0]+1,&fd,NULL,NULL,&timeout);
          
          if (signaled) {
            kill(pid,SIGTERM);
            MP3VolumeNow(getpid(),0,0);
	    break;
          }          
          
          if (waitpid(-1,NULL,WNOHANG)) break;
          
          if (FD_ISSET(pip[0],&fd)) {
            length = read(pip[0], buf, BUFFERSIZE);
            write( sock, buf, length );
          }                
      }
      close( sock );
      close( pip[0] );
      execl("/bin/true","/bin/true",NULL);
    }                                     
     
  } else {
//  WaitOnChild();
    return pid;
  }    
  return 0;
}


int MP3Stop(int id) {
  WaitOnChild();
  return kill(id,SIGTERM);
}

int WAVLoad(char *filename,double freq) {
  int sock,id,lh2,bits,length;
  int flags=ESD_STREAM | ESD_PLAY,rate;
  char s[100];
  char buf[BUFFERSIZE], header1[20], header2[1000];
  FILE *wav;
  struct stat wavstat;

  if ((wav=fopen(filename,"r"))==NULL) return -1;
  
  stat(filename,&wavstat);
  fread(header1,1,20,wav);
  lh2=*((int *)(header1+16));
  fread(header2,1,lh2,wav);
  if (*(short int *)header2!=1) return -1;
  if (*(short int *)(header2+2)==1) flags|=ESD_MONO;
  else if (*(short int *)(header2+2)==2) flags|=ESD_STEREO;
  else return -1;
  rate=(double)(*(unsigned int *)(header2+4))*freq;
  bits=(*(unsigned short int*)(header2+12)*8) / *(short int *)(header2+2);
  if (bits==8) flags |= ESD_BITS8;
  else if (bits==16) flags |= ESD_BITS16;
  else return -1;
  
  sock = esd_open_sound(esd_host);
  
  sprintf(s,"Sample[%s]",filename);
  id = esd_sample_cache(sock, flags, rate, wavstat.st_size-lh2-20,s);
  
  while ( (length=fread(buf,1,BUFFERSIZE,wav))>0 ) {
    write(sock,buf,length);
  }
  
  if (id!=esd_confirm_sample_cache(sock)) {
    esd_close(sock);  
    return -1;
  }    
  esd_close(sock);  
  return id;
}

int SamplePlay(int id, int l, int r) {
  int sock;
  sock = esd_open_sound(esd_host);
  esd_set_default_sample_pan(sock,id,l,r);    
  esd_sample_play(sock,id);
  esd_close(sock);  
  return 0;
}

int SampleLoop(int id, int l, int r) {
  int sock;
  sock = esd_open_sound(esd_host);
  esd_set_default_sample_pan(sock,id,l,r);    
  esd_sample_loop(sock,id);
  esd_close(sock);  
  return 0;
}

int SampleStop(int id,char now) {
  int sock;
  sock = esd_open_sound(esd_host);
//  if (now) esd_sample_kill(sock,id);
//    else esd_sample_stop(sock,id);
  esd_sample_stop(sock,id);
  esd_close(sock);  
  return 0;
}

int SampleUnload(int id) {
  int sock;
  sock = esd_open_sound(esd_host);
  esd_sample_free(sock,id);
  esd_close(sock);  
  return 0;
}

