program RescalePCX;
uses sysutils,
     crt;

type Tpal = array[0..1023] of byte;

     Timage1 = array[0..640*480-1] of byte;
     Pimage1 = ^Timage1;
     Timage2 = array[0..2560*1920-1] of byte;
     Pimage2 = ^Timage2;

type PCXheader = packed record
                      Identifier:byte;
                      Version:byte;
                      Encoding:byte;
                      BitsPerPixel:byte;
                      XStart:word;
                      YStart:word;
                      XEnd:word;
                      YEnd:word;
                      HorzRes:word;
                      VertRes:word;
                      Palette:array[0..47] of byte;
                      Reserved1:byte;
                      NumBitPlanes:byte;
                      BytesPerLine:word;
                      PaletteType:word;
                      HorzScreenSize:word;
                      VertScreenSize:word;
                      Reserved2:array[0..53] of byte;
                 end;


var pal:Tpal;
    i, j: integer;
    size: integer;
    w, h: integer;
    s, in_fn, out_fn: string;
    img2: PImage2;
    img1: PImage2;
    header: PCXheader;
    
    invert: boolean;

procedure LoadPCX(s:string; img: PImage2; var w,h: integer; var header:PCXheader; var pal:Tpal);
var fp:longint;
    f:file;
    data:array[0..2047] of byte;
    index:integer;
    colors:longint;

    i,count,linepaddingsize,pixelsperbyte:integer;
    x,y:longint;
    a,b:byte;

 function GetByte:byte;
 var done:integer;
 begin
     inc(index);
     if ((index=0) or (index>=1024)) then
     begin
          index:=0;
          BlockRead(f,data,1024,done);
     end;
     GetByte:=data[index];
 end;

begin
  assign(f,s);
  {$I-}
  reset(f,1);
  {$I+}
  if IOresult<>0 then exit;

  //main header
  BlockRead(f,header,128);
  if header.Identifier<>$0A then begin close(f); exit; end;

  header.XEnd:=header.XEnd-header.XStart+1;
  header.YEnd:=header.YEnd-header.YStart+1;
  colors:=1 shl (header.BitsPerPixel*header.NumBitPlanes);

  { len 256 farebne obrazky }
  if ((header.BitsPerPixel=8) and (header.NumBitPlanes=1)) then else begin close(f); exit; end;

{  ProcessColorMap(cols,o);
 }
   fp:=FilePos(f);
   seek(f,filesize(f)-769);
   blockread(f,pal,769);
   seek(f,fp);
   if data[0]=$0C then
   begin
{!}    {    for i:=0 to cols-1 do
	  setRGB(o+i,data[3*i+1],data[3*i+2],data[3*i+3]);}
   end;

{  ProcessData(cols,o);
 }
     PixelsPerByte:=8 div header.BitsPerPixel;
     LinePaddingSize:=(header.BytesPerLine*header.NumBitPlanes)*(8 div header.BitsPerPixel)-header.XEnd;
     LinePaddingSize:=LinePaddingSize div PixelsPerByte;
     w:=header.Xend;
     h:=header.Yend;

     index:=-1;
     y:=0;
     x:=0;
     Repeat
           a:=GetByte;
           If (a and $C0) = $C0 then
           Begin
                count:=a and $3F;
                b:=GetByte;
           End
           Else
           Begin
                count:=1;
                b:=a;
           End;

           For i:=1 to count do
           Begin
                img^[header.Xend*y+x]:=b;
                inc(x);
                If x*PixelsPerByte>=header.XEnd then
                Begin
                     inc(y);
                     x:=0;
                     inc(index,LinePaddingSize);
                     If y>=header.YEnd then break;
                End;
           End;
     Until y>=header.YEnd;

  close(f);
  colors:=colors;
end;

procedure ScalePCX(img1,img2:PImage2; size,w0,h0: longint; var w1,h1: integer; var pal:Tpal);
var v,x,y,i,j,xx,yy: longint;
begin
  w1:=w0 div size;
  h1:=h0 div size;
  
  for y:=0 to h1-1 do
  begin
   for x:=0 to w1-1 do
   begin
     v:=0;
     xx:=x*size;
     yy:=y*size;
     for i:=0 to size-1 do
      for j:=0 to size-1 do
       if img1^[w0*(yy+j)+xx+i]=img1^[0] then inc(v);
       
{     img2^[w1*y+x]:=3;  
     img2^[w1*y+x]:=img1^[w0*y+x];}
     img2^[w1*y+x]:=v;
   end;
  end;
   
  if invert then
  begin
    pal[1]:=255;  pal[2]:=255;  pal[3]:=255;
    pal[4]:=191;  pal[5]:=191;  pal[6]:=191;
    pal[7]:=127;  pal[8]:=127;  pal[9]:=127;
    pal[10]:=63;  pal[11]:=63;  pal[12]:=63;
    pal[13]:=0;  pal[14]:=0;  pal[15]:=0;
  end
  else
  begin
    pal[1]:=0;  pal[2]:=0;  pal[3]:=0;
    pal[4]:=63;  pal[5]:=63;  pal[6]:=63;
    pal[7]:=127;  pal[8]:=127;  pal[9]:=127;
    pal[10]:=191;  pal[11]:=191;  pal[12]:=191;
    pal[13]:=255;  pal[14]:=255;  pal[15]:=255;
  end;
end;


procedure SavePCX(out_fn: string; img: PImage2; w,h: integer; header:PCXheader; pal:Tpal);
var dbuf, ebuf: array[0..2560] of byte;
    f:file;
    index,scanindex,runcount: integer;
    runvalue: byte;
    y,x: integer;
    
begin
  header.Xstart:=0;
  header.Xend:=w-1;
  header.Ystart:=0;
  header.Yend:=h-1;
  header.bytesperline:=w;
  
  assign(f,out_fn);
  rewrite(f,1);
  blockwrite(f,header,128);
  
  for y:=0 to h-1 do  { encode each line }
  begin
    move(img^[w*y],dbuf[0],w);
    index:=0;
    scanindex:=0;
    while index<w do
    begin
      runvalue:=dbuf[index];
      runcount:=1;
      while ((runvalue=dbuf[index+runcount]) and (index+runcount<w) and (runcount<63)) do inc(runcount);
      
      if runcount>1 then
      begin
        ebuf[scanindex]:=runcount or $C0; inc(scanindex);
        ebuf[scanindex]:=runvalue; inc(scanindex);
      end
      else
      begin
        if dbuf[index]<64 then
	begin
	  ebuf[scanindex]:=runvalue;
	  inc(scanindex);
	end
	else
	begin
          ebuf[scanindex]:=runcount or $C0; inc(scanindex);
          ebuf[scanindex]:=runvalue; inc(scanindex);
	end;
      end;
      inc(index,runcount);
    end;
    blockwrite(f,ebuf,scanindex);
  end;

(*
  for y:=0 to h-1 do
   for x:=0 to w-1 do
   begin
     move(img^[w*y],dbuf[0],w);
     ebuf[0]:=$C1;
     ebuf[1]:=dbuf[x];
     blockwrite(f,ebuf[0],2);
   end;
*)
  
  blockwrite(f,pal,769);
  close(f);
end;


begin
  i:=0;
  size:=2;
  in_fn:='';
  out_fn:='';
  invert:=false;
  
  while i<paramcount do
  begin
    inc(i);
    s:=paramstr(i);
    if s='--size' then begin inc(i); size:=StrToInt(paramstr(i)); continue; end;
    if s='-o' then begin inc(i); out_fn:=paramstr(i); continue; end;
    if s='--invert' then begin invert:=true; continue; end;
    in_fn:=s;
  end;
  
  if in_fn='' then exit;
  
  new(img2);
  new(img1);
  LoadPCX(in_fn,img2,w,h,header,pal);
  ScalePCX(img2,img1,size,w,h,w,h,pal);
  SavePCX(out_fn,img1,w,h,header,pal);
  
  dispose(img2);
  dispose(img1);
end.
